/*  gentrex - generic tree expressions; gentrex based find(1)
    Copyright (C) 2015  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/gentrex
    Author: gentrex@igor2.repo.hu
*/

#define _XOPEN_SOURCE 600
#define _BSD_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/unistd.h>
#include <grp.h>
#include <pwd.h>
#include <dirent.h>
#include <regex.h>

#include "gentrex.h"
#include "helper_map_az.h"
#include "helper_intop.h"
#include "lex_helper.h"

#define GTX_SCRIPT_DEFAULT   "[]/*[]"
#define SHELL_CMD_DEFAULT    "echo \"$path\""
#define SHELL_DEFAULT        "/bin/sh"
#define PATH_DEFAULT         "."

static char *grcache[65535], *uscache[65535];

static char *get_group(gid_t gid)
{
	if ((gid > 65535) || (gid < 0))
		return "<invalid>";
	if (grcache[gid] == NULL) {
		struct group *g = getgrgid(gid);
		if (g != NULL)
			grcache[gid] = strclone(g->gr_name);
	}
	return grcache[gid];
}

static char *get_user(uid_t uid)
{
	if ((uid > 65535) || (uid < 0))
		return "<invalid>";
	if (uscache[uid] == NULL) {
		struct passwd *p = getpwuid(uid);
		if (p != NULL)
			uscache[uid] = strclone(p->pw_name);
	}
	return uscache[uid];
}

gentrex_map_az_t map, map_init;

typedef struct node_s node_t;

struct node_s {
	char *path;
	int pl;
	char *name;
	struct stat st;
	char *owner_name, *group_name;

	int cantlist;

	node_t *children, *next, *parent;
};

static node_t *node_alloc(const char *path, node_t *parent)
{
	node_t *nd;

	nd = malloc(sizeof(node_t));
	nd->pl = strlen(path);
	nd->path = malloc(nd->pl+1);
	memcpy(nd->path, path, nd->pl+1);
	nd->name = strrchr(nd->path, '/');
	if (nd->name == NULL)
		nd->name = nd->path;
	else
		nd->name++;
	stat(nd->path, &nd->st);

	nd->cantlist = 0;
	nd->children = NULL;
	nd->next = NULL;
	nd->parent = parent;

	nd->owner_name = get_user(nd->st.st_uid);
	nd->group_name = get_group(nd->st.st_gid);

	return nd;
}

static void ls(node_t *cwd)
{
	DIR *d;
	struct dirent *e;
	node_t *last_ch;

	if (cwd->cantlist)
		return;

	d = opendir(cwd->path);
	if (d == NULL) {
		cwd->cantlist = 1;
		return;
	}

	while((e = readdir(d)) != NULL) {
		node_t *ch;
		char path[8192];
		int fnl;

		if (e->d_name[0] == '.') {
			/* skip .. and . */
			if ((e->d_name[1] == '.') && (e->d_name[2] == '\0'))
				continue;
			if (e->d_name[1] == '\0')
				continue;
		}

		fnl = strlen(e->d_name);
		assert(fnl+cwd->pl+2 < sizeof(path));
		memcpy(path, cwd->path, cwd->pl);
		path[cwd->pl] = '/';
		memcpy(path+cwd->pl+1, e->d_name, fnl+1);

		ch = node_alloc(path, cwd);
		if (cwd->children != NULL) {
			last_ch->next = ch;
			last_ch = ch;
		}
		else
			cwd->children = last_ch = ch;
	}

	closedir(d);
}

static void *tree_child(void *user_ctx, void *node_)
{
	node_t *node = node_;
	if ((node->children == NULL) && (S_ISDIR(node->st.st_mode)))
		ls(node);
	return node->children;
}

static void *tree_parent(void *user_ctx, void *node_)
{
	node_t *node = node_;
	return node->parent;
}

static void *tree_next(void *user_ctx, void *node_)
{
	node_t *node = node_;
	return node->next;
}

typedef struct imatch_s {
	intop_t iop;
	long long int ref;
} imatch_t;


static void *op_cookie_alloc(void *user_tree_ctx, gentrex_op_t opt, const char *tagname, const char *str_or_pattern, const char *custom_operator)
{
	regex_t *rx;
	int upc;
	imatch_t *im;
	char *end;

	switch(opt) {
		case GTX_OP_EQ:
			break;
		case GTX_OP_C_OP:
			im = malloc(sizeof(imatch_t));
			im->ref = strtoll(str_or_pattern, &end, 0);
			switch(*end) {
				case 'k': end++; im->ref *= 1024LL; break;
				case 'm': end++; im->ref *= 1024LL*1024LL; break;
				case 'g': end++; im->ref *= 1024LL*1024LL*1024LL; break;
				case '\0': break;
			}
			if (*end != '\0') {
				fprintf(stderr, "Invalid integer: '%s'\n", str_or_pattern);
				exit(1);
			}
			im->iop = intop_parse(custom_operator);
			if (im->iop == INTOP_INVALID) {
				fprintf(stderr, "Invalid integer operator: '%s'\n", custom_operator);
				exit(1);
			}
			return im;
		case GTX_OP_MATCH:
			upc = ((*tagname >= 'A') && (*tagname <= 'Z')) ? REG_ICASE : 0;
			rx = malloc(sizeof(regex_t));
			if (regcomp(rx, str_or_pattern, REG_EXTENDED | upc | REG_NOSUB) != 0) {
				fprintf(stderr, "Invalid regex '%s'\n", str_or_pattern);
				exit(1);
			}
			return rx;
	}
	return NULL;
}

static void op_cookie_free(void *user_tree_ctx, gentrex_op_t opt, const char *tagname, const char *str_or_pattern, void *op_cookie, const char *custom_operator)
{
	switch(opt) {
		case GTX_OP_EQ:
			break;
		case GTX_OP_C_OP:
			free(op_cookie);
			break;
		case GTX_OP_MATCH:
			regfree((regex_t *)op_cookie);
			free(op_cookie);
			break;
	}
}

static int tag_eq(void *user_ctx, void *node_, void *op_cookie, const char *tagname, const char *str)
{
	node_t *node = node_;
	int ival = atoi(str);
	switch(*tagname) {
		case 'n': return !strcmp(node->name, str);
		case 'N': return !strcasecmp(node->name, str);
		case 'p': return !strcmp(node->path, str);
		case 'P': return !strcasecmp(node->path, str);
		case 'o': return !strcmp(node->owner_name, str);
		case 'O': return !strcasecmp(node->owner_name, str);
		case 'g': return !strcmp(node->group_name, str);
		case 'G': return !strcasecmp(node->group_name, str);

		case 's': return node->st.st_size == ival;
		case 'l': return node->st.st_nlink == ival;

	}
	return 0;
}

int tag_match(void *user_tree_ctx, void *node_, void *op_cookie, const char *tagname, const char *pattern)
{
	node_t *node = (node_t *)node_;
	regex_t *rx = (regex_t *)op_cookie;
	char sval[64];

	switch(*tagname) {
		case 'n':
		case 'N': return !regexec(rx, node->name, 0, NULL, 0);
		case 'p':
		case 'P': return !regexec(rx, node->path, 0, NULL, 0);
		case 'o':
		case 'O': return !regexec(rx, node->owner_name, 0, NULL, 0);
		case 'g':
		case 'G': return !regexec(rx, node->group_name, 0, NULL, 0);
		case 's': sprintf(sval, "%ld", node->st.st_size); return !regexec(rx, sval, 0, NULL, 0);
	}
	return 0;
}

int tag_c_op(void *user_tree_ctx, void *node_, void *op_cookie, const char *tagname, const char *str, const char *c_op)
{
	long long int ival;
	node_t *node = (node_t *)node_;
	imatch_t *im = (imatch_t *)op_cookie;

	switch(*tagname) {
		case 's': ival = node->st.st_size; break;
		case 'u': ival = node->st.st_uid; break;
		case 'q': ival = node->st.st_gid; break;
		case 'l': ival = node->st.st_nlink; break;
	}

	intop_cmp(ival, im->iop, im->ref, return 0);

	return 0;
}

static int need_unset[26];
static void fprint_nodevars(FILE *fcmd, char *prefix, node_t *node, int idx)
{
	if (node != NULL) {
		fprintf(fcmd, "%sname='%s';", prefix, node->name);
		fprintf(fcmd, "%spath='%s';", prefix, node->path);
		fprintf(fcmd, "%ssize='%ld';", prefix, node->st.st_size);
		fprintf(fcmd, "%suid='%d';", prefix, node->st.st_uid);
		fprintf(fcmd, "%sqid='%d';", prefix, node->st.st_gid);
		fprintf(fcmd, "%slink='%d';", prefix, node->st.st_nlink);
		fprintf(fcmd, "%sowner='%s';", prefix, node->owner_name);
		fprintf(fcmd, "%sgroup='%s';", prefix, node->group_name);
		if (idx >= 0)
			need_unset[idx] = 1;
	}
	else if ((idx < 0) || (need_unset[idx])) {
		fprintf(fcmd, "unset %sname", prefix);
		fprintf(fcmd, " %spath", prefix);
		fprintf(fcmd, " %ssize", prefix);
		fprintf(fcmd, " %suid", prefix);
		fprintf(fcmd, " %sqid", prefix);
		fprintf(fcmd, " %slink", prefix);
		fprintf(fcmd, " %sowner", prefix);
		fprintf(fcmd, " %sgroup", prefix);
		fprintf(fcmd, ";");
		need_unset[idx] = 0;
	}
}

void help(const char *prg)
{
	printf(
		"\ngtxfind - search files recursively using gentrex patter\n\n"
		"Usage:\n"
		"%s [-g gtx] [-e cmd] [-p path] [-i init] [-u uninit] [-s shell] [-h]\n"
		"\n"
		" -g gtx      gentrex script (default:" GTX_SCRIPT_DEFAULT ")\n"
		" -e cmd      shell command to run on each result (default:" SHELL_CMD_DEFAULT ")\n"
		" -p path     path to start the search from (default:" PATH_DEFAULT ")\n"
		"", prg);

	printf(
		" -i init     shell command to run first (default: empty)\n"
		" -u uninit   shell command to run last (default: empty)\n"
		" -s shell    path to a POSIX shell executable (default: " SHELL_DEFAULT ")\n"
		" -h          this help screen\n"
		"\n"
	);
	printf(
		"Nodes of the tree are filesystem objects with the following fields:\n"
		" name       base name\n"
		" path       path relative to -p\n"
		" size       size in bytes\n"
		" uid        owner user ID (int)\n"
		" qid        owner group ID (int)\n"
		" owner      owner user name\n"
		" group      owner group name\n"
		"\n"
	);

	printf(
		"When referenced from tag_expr ([]), if field name is uppercase, string\n"
		"comparison or regex match is performed case insensitive. The ~ operator\n"
		"performs extended posix regex match. Map names are lowercase single\n"
		"letters (a..z). The map is reset before each gentrex iteration.\n"
		"For each result, all non-empty map fields are copied into shell variables\n"
		"called: $map_field, e.g. \"$p_uid\" for map variable p's uid. The result node"
		"is copied into unprefixed variables (e.g. \"$uid\", \"$path\", etc.)\n"
		"\n"
		);
}

#define asg_once(target) \
do { \
	if (target != NULL) { \
		fprintf(stderr, "Error: can't specify multiple %s values\n", argv[n]); \
		exit(1); \
	} \
	target = arg; \
	n++; \
} while(0)

int main(int argc, char *argv[])
{
	node_t *root;
	gentrex_state_t *st;
	gentrex_api_t api;
	FILE *fcmd;
	const char *err;
	char vp[3];
	char *cmd, *arg;
	int n;

	char *gtx_script = NULL;
	char *shell_cmd = NULL;
	char *shell_init = NULL;
	char *shell_uninit = NULL;
	char *shell = NULL;
	char *path = NULL;

	for(n = 1; n < argc; n++) {
		cmd = argv[n];
		arg = argv[n+1];
		while(*cmd == '-') cmd++;
		switch(*cmd) {
			case 'g': asg_once(gtx_script); break;
			case 'e': asg_once(shell_cmd); break;
			case 's': asg_once(shell); break;
			case 'i': asg_once(shell_init); break;
			case 'u': asg_once(shell_uninit); break;
			case 'p': asg_once(path); break;
			case 'h': help(argv[0]); exit(0);
			default:
				fprintf(stderr, "Invalid command line argument '%s'\n", argv[n]);
				exit(1);
		}
	}

	if (gtx_script == NULL)  gtx_script = GTX_SCRIPT_DEFAULT;
	if (shell_cmd  == NULL)  shell_cmd  = SHELL_CMD_DEFAULT;
	if (shell      == NULL)  shell      = SHELL_DEFAULT;
	if (path       == NULL)  path       = PATH_DEFAULT;


	fcmd=popen(shell, "w");
	if (fcmd == NULL) {
		fprintf(stderr, "Failed to execute shell %s\n", shell);
	}

	root = node_alloc(path, NULL);
	ls(root);

	/* set up the API using the example tree and the simplistic a-z map */
	memset(&api, 0, sizeof(api));
	api.user_map_ctx = &map;
	api.map_get = gentrex_map_az_get;
	api.map_set = gentrex_map_az_set;

	api.next   = tree_next;
	api.child  = tree_child;
	api.parent = tree_parent;

	api.op_cookie_alloc = op_cookie_alloc;
	api.op_cookie_free  = op_cookie_free;
	api.tag_eq          = tag_eq;
	api.tag_match       = tag_match;
	api.tag_c_op        = tag_c_op;

	st = gentrex_alloc(&api, gtx_script, &err);
	if (st == NULL) {
		fprintf(stderr, "gentrex error: %s\n", err);
		return 1;
	}

	memset(&map_init, 0, sizeof(map_init));

	if (shell_init != NULL)
		fprintf(fcmd, "%s\n", shell_init);


	vp[1] = '_';
	vp[2] = '\0';
	/* iterate over all matches */
	for(;;) {
		node_t *res;
		int n;

		/* restore the initial map for each subsequent query */
		memcpy(&map, &map_init, sizeof(map));

		/* eval the expression, get the next match */
		res = gentrex_execute(st, root);
		if (res == NULL)
			break;

		/* print all variables from the map that are requested with -p on the CLI */
		for(n = 0; n < 26; n++) {
			vp[0] = n + 'a';
			fprint_nodevars(fcmd, vp, ((node_t *)map.map[n].straight), n);
		}

		fprint_nodevars(fcmd, "", res, -1);
		fprintf(fcmd, "%s\n", shell_cmd);

		fprintf(fcmd, "\n");
	}

	if (shell_uninit != NULL)
		fprintf(fcmd, "%s\n", shell_uninit);

	pclose(fcmd);

	return 0;
}
